@@ -1,9 +1,9 @@
 /*
  *  aespipe.c
  *
- *  Written by Jari Ruusu, July 7 2004
+ *  Written by Jari Ruusu, February 18 2007
  *
- *  Copyright 2002,2003,2004 by Jari Ruusu.
+ *  Copyright 2002-2007 by Jari Ruusu.
  *  Redistribution of this file is permitted under the GNU Public License.
  *
  *  AES encrypting or decrypting "pipe", reads from stdin, writes to stdout
@@ -17,6 +17,7 @@
 #include <stdlib.h>
 #include <pwd.h>
 #include <sys/types.h>
+#include <signal.h>
 #if HAVE_SYS_WAIT_H
 # include <sys/wait.h>
 #endif
@@ -60,6 +61,7 @@ char            *passSeedString = (char *)0;
 int             passAskTwice = 0;
 char            *gpgKeyFile = (char *)0;
 char            *gpgHomeDir = (char *)0;
+char            *clearTextKeyFile = (char *)0;
 char            *passIterThousands = (char *)0;
 int             complainWriteErr = 1;
 unsigned int    waitSeconds = 0;
@@ -69,9 +71,10 @@ union {
     u_int32_t       w[2048];
     unsigned char   b[8192];
 } buf;
-int             multiKeyMode = 0;
-char            *multiKeyPass[64];
+int             multiKeyMode = 0; /* 0=single-key 64=multi-key-v2 65=multi-key-v3 */
+char            *multiKeyPass[66];
 aes_context     multiKeyCtx[64];
+u_int32_t       partialMD5[4];
 
 int rd_wr_retry(int fd, char *buf, int cnt, int w)
 {
@@ -126,12 +129,20 @@ char *get_FD_pass(int fd)
     return p;
 }
 
+static void warnAboutBadKeyData(int x)
+{
+    if((x > 1) && (x != 64) && (x != 65)) {
+        fprintf(stderr, "Warning: Unknown key data format - using it anyway\n");
+    }
+}
+        
 char *do_GPG_pipe(char *pass)
 {
     int     x, pfdi[2], pfdo[2];
     char    str[10], *a[16], *e[2], *h;
     pid_t   gpid;
     struct passwd *p;
+    void    *oldSigPipeHandler;
 
     if(gpgHomeDir && gpgHomeDir[0]) {
         h = gpgHomeDir;
@@ -189,8 +200,7 @@ char *do_GPG_pipe(char *pass)
             a[x++] = "--homedir";
             a[x++] = gpgHomeDir;
         }
-        a[x++] = "--options";
-        a[x++] = "/dev/null";
+        a[x++] = "--no-options";
         a[x++] = "--quiet";
         a[x++] = "--batch";
         a[x++] = "--no-tty";
@@ -230,12 +240,17 @@ char *do_GPG_pipe(char *pass)
     }
 
     x = strlen(pass);
+
+    /* ignore possible SIGPIPE signal while writing to gpg */
+    oldSigPipeHandler = signal(SIGPIPE, SIG_IGN);
     rd_wr_retry(pfdi[1], pass, x, 1);
     rd_wr_retry(pfdi[1], "\n", 1, 1);
+    if(oldSigPipeHandler != SIG_ERR) signal(SIGPIPE, oldSigPipeHandler);
+
     close(pfdi[1]);
     memset(pass, 0, x);
     x = 0;
-    while(x < 64) {
+    while(x < 66) {
         multiKeyPass[x] = get_FD_pass(pfdo[0]);
         if(!multiKeyPass[x]) {
             /* mem alloc failed - abort */
@@ -245,8 +260,11 @@ char *do_GPG_pipe(char *pass)
         if(strlen(multiKeyPass[x]) < AESPIPE_PASSWORD_MIN_LENGTH) break;
         x++;
     }
+    warnAboutBadKeyData(x);
+    if(x >= 65)
+        multiKeyMode = 65;
     if(x == 64)
-        multiKeyMode = 1;
+        multiKeyMode = 64;
     close(pfdo[0]);
     waitpid(gpid, &x, 0);
     if(!multiKeyPass[0]) goto nomem1;
@@ -280,13 +298,44 @@ char *getPass(char *prompt)
 char *sGetPass(int minLen)
 {
     char *p, *s, *seed;
-    int i, ask2;
+    int i, ask2, close_psw_fd = 0;
 
     if(passFDnumber < 0) {
+        if(clearTextKeyFile) {
+            if((passFDnumber = open(clearTextKeyFile, O_RDONLY)) == -1) {
+                fprintf(stderr, "Error: unable to open %s for reading\n", clearTextKeyFile);
+                return NULL;
+            }
+            close_psw_fd = 1;
+            goto contReadFrom_psw;
+        }
         p = getPass("Password: ");
         ask2 = passAskTwice;
     } else {
-        p = get_FD_pass(passFDnumber);
+        contReadFrom_psw:
+        if(gpgKeyFile && gpgKeyFile[0]) {
+            p = get_FD_pass(passFDnumber);
+            if(close_psw_fd) close(passFDnumber);
+        } else {
+            int x = 0;
+            while(x < 66) {
+                multiKeyPass[x] = get_FD_pass(passFDnumber);
+                if(!multiKeyPass[x]) goto nomem;
+                if(strlen(multiKeyPass[x]) < AESPIPE_PASSWORD_MIN_LENGTH) break;
+                x++;
+            }
+            if(close_psw_fd) close(passFDnumber);
+            warnAboutBadKeyData(x);
+            if(x >= 65) {
+                multiKeyMode = 65;
+                return multiKeyPass[0];
+            }
+            if(x == 64) {
+                multiKeyMode = 64;
+                return multiKeyPass[0];
+            }
+            p = multiKeyPass[0];
+        }
         ask2 = 0;
     }
     if(!p) goto nomem;
@@ -340,6 +389,49 @@ char *sGetPass(int minLen)
     return(s);
 }
 
+/* obsolete */
+void unhashed1_hash_buffer(unsigned char *keyStr, int ile, unsigned char *keyBuf, int bufSize) {
+    int x, y, z, cnt = ile;
+    unsigned char *kp;
+
+    memset(keyBuf, 0, bufSize);
+    kp = keyStr;
+
+    for (x = 0; x < (bufSize * 8); x += 6) {
+        y = *kp++;
+
+        if (--cnt <= 0) {
+            kp = keyStr;
+            cnt = ile;
+        }
+
+        if ((y >= '0') && (y <= '9')) y -= '0';
+        else if((y >= 'A') && (y <= 'Z')) y -= ('A' - 10);
+        else if((y >= 'a') && (y <= 'z')) y -= ('a' - 36);
+        else if((y == '.') || (y == '/')) y += (62 - '.');
+        else y &= 63;
+
+        z = x >> 3;
+
+        if (z < bufSize) {
+            keyBuf[z] |= y << (x & 7);
+        }
+
+        z++;
+
+        if (z < bufSize) {
+            keyBuf[z] |= y >> (8 - (x & 7));
+        }
+    }
+}
+
+/* obsolete */
+void unhashed2_hash_buffer(unsigned char *keyStr, int ile, unsigned char *keyBuf, int bufSize) {
+    memset(keyBuf, 0, bufSize);
+    strncpy((char *)keyBuf, (char *)keyStr, bufSize - 1);
+    keyBuf[bufSize - 1] = 0;
+}
+
 void rmd160HashTwiceWithA(unsigned char *ib, int ile, unsigned char *ob, int ole)
 {
     unsigned char tmpBuf[20 + 20];
@@ -358,16 +450,11 @@ void rmd160HashTwiceWithA(unsigned char *ib, int ile, unsigned char *ob, int ole
     memset(pwdCopy, 0, sizeof(pwdCopy));
 }
 
-void compute_md5_iv(u_int32_t *data)
+void compute_md5_iv_v3(u_int32_t *data)
 {
     int         x, y, e;
     u_int32_t   buf[16];
 
-    iv[0] = 0x67452301;
-    iv[1] = 0xefcdab89;
-    iv[2] = 0x98badcfe;
-    iv[3] = 0x10325476;
-
     y = 7;
     e = 16;
     do {
@@ -379,6 +466,8 @@ void compute_md5_iv(u_int32_t *data)
             buf[12] = devSect0;
             buf[13] = (devSect1 & 0xFFFFFF) | 0x80000000;
             /* 4024 bits == 31 * 128 bit plaintext blocks + 56 bits of sector number */
+            /* For version 3 on-disk format this really should be 4536 bits, but can't be */
+            /* changed without breaking compatibility. V3 uses MD5-with-wrong-length IV */
             buf[14] = 4024;
             buf[15] = 0;
         }
@@ -410,7 +499,8 @@ void doEncryptMD5ivCBC(int cnt)
 
     do {
         a = &multiKeyCtx[devSect0 & 0x3F];
-        compute_md5_iv((u_int32_t *)(datap+16));
+        memcpy(&iv[0], &partialMD5[0], 16);
+        compute_md5_iv_v3((u_int32_t *)(datap+16));
         x = 32;
         do {
             iv[0] ^= *((u_int32_t *)(&datap[ 0]));
@@ -447,7 +537,8 @@ void doDecryptMD5ivCBC(int cnt)
             datap += 16;
         } while(--x);
         datap -= 512;
-        compute_md5_iv((u_int32_t *)(datap+16));
+        memcpy(&iv[0], &partialMD5[0], 16);
+        compute_md5_iv_v3((u_int32_t *)(datap+16));
         aes_decrypt(a, datap, datap);
         *((u_int32_t *)(&datap[ 0])) ^= iv[0];
         *((u_int32_t *)(&datap[ 4])) ^= iv[1];
@@ -512,8 +603,8 @@ int main(int argc, char **argv)
     int x, encrypt = 1, bits, ret;
     void (*hf)(unsigned char *, int, unsigned char *, int);
     union {
-        u_int32_t     hw[8]; /* needed for 4 byte alignment of hb[] */
-        unsigned char hb[32];
+        u_int32_t     hw[16];
+        unsigned char hb[64];
     } hbu;
     char *pass, *hfn = (char *)0, *efn = (char *)0;
     unsigned int y;
@@ -531,11 +622,12 @@ int main(int argc, char **argv)
         if(!strcmp(*argv, "-") || (**argv != '-')) {
             usage:
             fprintf(stderr, "usage: %s [options] <inputfile >outputfile\n"
-                            "version 2.2d  Copyright (c) 2002-2004 Jari Ruusu, (c) 2001 Dr Brian Gladman\n"
+                            "version 2.3d  Copyright (c) 2002-2007 Jari Ruusu, (c) 2001 Dr Brian Gladman\n"
                             "options:  -e aes128|aes192|aes256          =  set key length\n"
                             "          -H sha256|sha384|sha512|rmd160   =  set password hash function\n"
                             "          -d         =  decrypt\n"
                             "          -p num     =  read password from file descriptor num\n"
+                            "          -P file    =  read password from file\n"
                             "          -S pseed   =  set password seed\n"
                             "          -T         =  ask password twice\n"
                             "          -q         =  don't complain about write errors\n"
@@ -564,6 +656,10 @@ int main(int argc, char **argv)
                     if(!(*++(*argv) || (--argc && *++argv))) goto usage;
                     if(sscanf(*argv, "%d", &passFDnumber) != 1) goto usage;
                     goto nextArg;
+                case 'P':
+                    if(!(*++(*argv) || (--argc && *++argv))) goto usage;
+                    clearTextKeyFile = *argv;
+                    goto nextArg;
                 case 'S':
                     if(!(*++(*argv) || (--argc && *++argv))) goto usage;
                     passSeedString = *argv;
@@ -612,7 +708,7 @@ int main(int argc, char **argv)
         } else if(!strcasecmp(efn, "aes192")) {
             bits = 192;
             hf = sha384_hash_buffer;
-        } else if(strcasecmp(efn, "aes128")) {
+        } else if(strcasecmp(efn, "aes128") && strcasecmp(efn, "aes")) {
             goto usage;
         }
     }
@@ -621,6 +717,8 @@ int main(int argc, char **argv)
         else if(!strcasecmp(hfn, "sha384")) hf = sha384_hash_buffer;
         else if(!strcasecmp(hfn, "sha512")) hf = sha512_hash_buffer;
         else if(!strcasecmp(hfn, "rmd160")) hf = rmd160HashTwiceWithA;
+        else if(!strcasecmp(hfn, "unhashed1")) hf = unhashed1_hash_buffer;
+        else if(!strcasecmp(hfn, "unhashed2")) hf = unhashed2_hash_buffer;
         else goto usage;
     }
 
@@ -628,26 +726,51 @@ int main(int argc, char **argv)
     pass = sGetPass((hf == rmd160HashTwiceWithA) ? 1 : AESPIPE_PASSWORD_MIN_LENGTH);
     if(!pass) exit(1);
     x = strlen(pass);
-    (*hf)((unsigned char *)pass, x, &hbu.hb[0], sizeof(hbu.hb));
+    if(hf == unhashed1_hash_buffer) { /* obsolete compat */
+        bits = 128;
+        if(x >= 32) bits = 192;
+        if(x >= 43) bits = 256;
+    }
+    (*hf)((unsigned char *)pass, x, &hbu.hb[0], 32);
     if(multiKeyMode) {
         int r = 0, t;
-        while(r < 64) {
+        partialMD5[0] = 0x67452301;
+        partialMD5[1] = 0xefcdab89;
+        partialMD5[2] = 0x98badcfe;
+        partialMD5[3] = 0x10325476;
+        while(r < multiKeyMode) {
             t = strlen(multiKeyPass[r]);
-            (*hf)(multiKeyPass[r], t, &hbu.hb[0], sizeof(hbu.hb));
+            (*hf)(multiKeyPass[r], t, &hbu.hb[0], 32);
             memset(multiKeyPass[r], 0, t);
             /*
              * MultiKeyMode uses md5 IV. One key mode uses sector IV. Sector IV
-             * and md5 IV are computed differently. This first key byte XOR with
-             * 0x55 is needed to cause complete decrypt failure in cases where
-             * data is encrypted with sector IV and decrypted with md5 IV or
-             * vice versa. If identical key was used but only IV was computed
-             * differently, only first plaintext block of 512 byte CBC chain
-             * would decrypt incorrectly and rest would decrypt correctly.
+             * and md5 IV v2 and v3 are all computed differently. This first key
+             * byte XOR with 0x55/0xF4 is needed to cause complete decrypt failure
+             * in cases where data is encrypted with one type of IV and decrypted
+             * with another type IV. If identical key was used but only IV was
+             * computed differently, only first plaintext block of 512 byte CBC
+             * chain would decrypt incorrectly and rest would decrypt correctly.
              * Partially correct decryption is dangerous. Decrypting all blocks
              * incorrectly is safer because file system mount will simply fail.
              */
-            hbu.hb[0] ^= 0x55;
-            aes_set_key(&multiKeyCtx[r], &hbu.hb[0], bits, 0);
+            if(multiKeyMode == 65) {
+                hbu.hb[0] ^= 0xF4; /* version 3 */
+            } else {
+                hbu.hb[0] ^= 0x55; /* version 2 */
+            }
+            if(r < 64) {
+                aes_set_key(&multiKeyCtx[r], &hbu.hb[0], bits, 0);
+            } else {
+                /* only first 128 bits of iv-key is used */
+#if WORDS_BIGENDIAN
+                hbu.hw[0] = xcpu_to_le32(hbu.hw[0]);
+                hbu.hw[1] = xcpu_to_le32(hbu.hw[1]);
+                hbu.hw[2] = xcpu_to_le32(hbu.hw[2]);
+                hbu.hw[3] = xcpu_to_le32(hbu.hw[3]);
+#endif
+                memset(&hbu.hb[16], 0, 48);
+                md5_transform_CPUbyteorder(&partialMD5[0], &hbu.hw[0]);
+            }
             r++;
         }
     } else if(passIterThousands) {
@@ -711,5 +834,6 @@ int main(int argc, char **argv)
     memset(&multiKeyCtx[0], 0, sizeof(multiKeyCtx));
     memset(&iv[0], 0, sizeof(iv));
     memset(&buf.b[0], 0, sizeof(buf));
+    memset(&partialMD5[0], 0, sizeof(partialMD5));
     exit(ret);
 }
